package Kakuro;

import choco.integer.IntDomainVar;
import choco.AbstractProblem;
import choco.Problem;
import choco.ContradictionException;
import choco.util.IntIterator;

import java.util.*;

/**
 * A class representing a Kakuro Problem and providing
 * resolution services
 **/
public class KakuroProblem {

	/**
	 * 	an identifier to the group of problems this one belongs
	 */
	protected String kind;

	/**
	 * 	its number among its group
	 */
	protected int number;

	/**
	 * store the resolution time for statistics
	 */
	protected long resolutionTime;

	/**
	 * 	representation of the matrix (with null entries)
	 */
	protected Cell[][] kmatrix;

	/**
	 * 	the list of all sequences of cells that have to constrained by sum+alldiff
	 */
	protected List<Sequence> seqs;

	/**
	 * store all possible sumalldiff subproblems (with null entries)
	 * fro each value of the sum (1 to 45) and size of the sequence (1 to 9)
	 * Dim : 45 * 9 
	 */
	protected SumAlldiffPb[][] subpbs_mat;

	/**
	 * The choco entity representing the problem
	 */
	protected AbstractProblem pb;

	protected boolean solveWithoutShaving = false;
	protected boolean solveWithoutSearch = false;


	public KakuroProblem(String kind, int number) {
		this.pb = new Problem();
		this.kind = kind;
		this.number = number;
		subpbs_mat = new SumAlldiffPb[46][10];
		seqs = new LinkedList<Sequence>();
	}

	public SumAlldiffPb setSumAlldiffPb(int nbvar, int sum) {
		if (subpbs_mat[sum][nbvar] == null)
			subpbs_mat[sum][nbvar] = new SumAlldiffPb(nbvar, sum);
		return subpbs_mat[sum][nbvar];
	}

	public void createMatrix(int nblig, int nbcol) {
		kmatrix = new Cell[nblig][nbcol];
	}

	public void setCell(Cell c) {
		kmatrix[c.i][c.j] = c;
	}

	public Cell getCell(int i, int j) {
		return kmatrix[i][j];
	}

	public void addSequence(Sequence sq) {
		seqs.add(sq);
	}


	public String getKind() {
		return kind;
	}

	public int getNumber() {
		return number;
	}

	public long getResolutionTime() {
		return resolutionTime;
	}

	public boolean isSolveWithoutSearch() {
		return solveWithoutSearch;
	}

	public boolean isSolveWithoutShaving() {
		return solveWithoutShaving;
	}

	/**
	 * Main solve method of the puzzle :
	 * 1- compute accurate bounds for each kind of sumAllDiff
	 * (identified by the lenght of the sequence and value of the sum)
	 * 2- then create the choco variables for the cells with accurate bounds
	 * computed previously for each sumAllDiff sub problems
	 * 3- post the sumAllDiff as the conjonction of sum and alldiff and propagate
	 * them to start simplifying the problem
	 * 4- Sort the sequences from the smallest to largest abd post the sumAllDiff
	 * represented this time as a regular and propagate it immediatly to simplify
	 * the problem at each step and speed up the generation of the next regular.
	 * 5- Apply a step of shaving if all the cells are not known yet
	 * 6- Call a solve there still exist a non intantiated cell in the grid.
	 * (this is very unlikely to occur and pyzzles are solved by the previous reasonning
	 * only in practice) 
	 */
	public void solveKakuro() {
		try {

			resolutionTime = System.currentTimeMillis();
			//create subproblems bounds
			for (int i = 0; i < subpbs_mat.length; i++) {
				for (int j = 0; j < subpbs_mat[0].length; j++) {
					SumAlldiffPb spbs = subpbs_mat[i][j];
					if (spbs != null)
						spbs.computeBounds();
				}
			}

			//create variable with accurate bounds obtain from
			//the resolution of subproblems
			for (int i = 0; i < kmatrix.length; i++) {
				for (int j = 0; j < kmatrix[0].length; j++) {
					Cell c = kmatrix[i][j];
					if (c != null) c.createVariable(pb);
				}
			}
			//post naive constraints
			for (Sequence s : seqs) {
				s.computeVars(kmatrix);
				s.postNaiveSequence(pb);
			}

			//further reduce domains before going to costly reasonnings
			pb.propagate();

			Collections.sort(seqs, new SequenceComparator());
			//post complex reasonnings
			for (Sequence s : seqs) {
				s.postComplexSum_AllDiff(pb);
				//further reduce domains before going to costly reasonnings
				pb.propagate();
			}


			//Shaving step if not yet solved
			if (!isSol()) {
				shavingWihoutFixPoint();
			} else {
				solveWithoutShaving = true;
			}

			//search otherwise (should not happen in practice)
			if (!isSol()) {
				System.out.println("Search is needed on this puzzle ?");
				pb.solve();
			} else solveWithoutSearch = true;
			resolutionTime = (System.currentTimeMillis() - resolutionTime);
		} catch (ContradictionException e) {
			e.printStackTrace();
			throw new Error("error: Kakuro should be consistent");
		}
	}

	/**
	 * Check if all the cells of the puzzle are instantiated
	 * @return
	 */
	public boolean isSol() {
		for (int i = 0; i < pb.getNbIntVars(); i++) {
			IntDomainVar v = (IntDomainVar) pb.getIntVar(i);
			if (!v.isInstantiated()) return false;
		}
		return true;
	}


	/**
	 * Apply a step of shaving on the remaining non instantiated variable
	 */
	public void shavingWihoutFixPoint() {
		pb.worldPush();
		for (int i = 0; i < pb.getNbIntVars(); i++) {
			IntDomainVar v = (IntDomainVar) pb.getIntVar(i);
			if (!v.isInstantiated()) {
				IntIterator it = v.getDomain().getIterator();
				while (it.hasNext()) {
					int val = it.next();
					boolean contradictionCatched = false;
					pb.worldPush();
					try {
						v.setVal(val);
						pb.propagate();
					} catch (ContradictionException e) {
						contradictionCatched = true;
					}
					pb.worldPop();
					if (contradictionCatched) {
						try {
							v.remVal(val);
							pb.propagate();
						} catch (ContradictionException e) {
							throw new Error("Shaving: Kakuro should be consistent");
						}
					}
				}
			}
		}
	}

	/**
	 * Print the solution contained in the choco variables
	 */
	public void printSolution() {
		for (int i = 0; i < kmatrix.length; i++) {
			for (int j = 0; j < kmatrix[0].length; j++)
				System.out.print(kmatrix[i][j] == null ? " " : kmatrix[i][j]);
			System.out.println();
		}
	}

}
