package Kakuro.parser;

import Kakuro.KakuroProblem;

import java.io.IOException;
import java.util.List;
import java.util.LinkedList;

/**
 * A class to deal with a set of problems define as prolog facts.
 * It will read an .ecl file containing a set of problems, solve
 * each of them in turn and output some statistics on their resolution
 **/
public class Benchmark {

	public String file;

	//The statistic are gathered by "kind" of problem defined in the ecl file
	public List<Statistic> stats;
	
	public Statistic all;

	/**
	 * Create a benchmark suite from a given ".ecl" file
	 * @param file
	 */
	public Benchmark(String file) {
		this.file = file;
		stats = new LinkedList<Statistic>();
		all = new Statistic("All");
	}

	/**
	 * Main lop to launch to resolution of all problems
	 */
	public void solveAll() {
		try {

			Reader reader = new Reader(file);
			KakuroProblem kpb = reader.loadNext();

			do{

				kpb.solveKakuro();
				kpb.printSolution();
				addToStatistic(kpb);
				kpb = reader.loadNext();

			} while(kpb != null);

			printLatexTable();
		
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public boolean addToStatistic(KakuroProblem pb) {
		all.addProblem(pb);
		for (Statistic s: stats) {
			if (s.kind.equals(pb.getKind())) {
				s.addProblem(pb);
				return false;
			}
		}
		Statistic s = new Statistic(pb.getKind());
		s.addProblem(pb);
		stats.add(s);
		return true;
	}

	public void printLatexTable() {
		all.printHeadLine();
		for (Statistic s: stats) {
			s.finalizeStats();
			s.printLatexLine();
		}
		all.finalizeStats();
		all.printLatexLine();
	}



	public class Statistic {

		public String kind;
		public int numberOfPbs = 0;
		public float percentSolvedBySetup = 0;
		public float percentSolvedByShaving = 0;
		public float avgtime = 0;
		public float maxtime = -1;

		public Statistic(String k) {
			kind = k;
		}

		public void addProblem(KakuroProblem pb) {
			numberOfPbs ++;
			if (pb.isSolveWithoutShaving()) {
				percentSolvedBySetup += 1;
			}
			if (pb.isSolveWithoutSearch()) {
				percentSolvedByShaving += 1;
			}
			avgtime += pb.getResolutionTime();
			if (pb.getResolutionTime() > maxtime) {
				maxtime = pb.getResolutionTime();
			}
		}

		public void finalizeStats() {
			percentSolvedBySetup = percentSolvedBySetup / numberOfPbs * 100;
			percentSolvedByShaving = percentSolvedByShaving / numberOfPbs * 100;
			avgtime = avgtime / numberOfPbs;
		}

		public void printHeadLine() {
			System.out.println("kind  & numberOfPbs & %SolvedByPropag & %SolvedByShaving & avg_time  & max_time  \\\\");
		}

		public void printLatexLine() {
			System.out.println(kind + " & " + numberOfPbs + " & " + round2digits(percentSolvedBySetup) + " & " +
							   round2digits(percentSolvedByShaving) + " & " + roundInSeconds(avgtime) + " & " + roundInSeconds(maxtime) + " \\\\");
		}

		public float roundInSeconds(float f) {
		    return	 Math.round(f / 10f) / 100f;
		}

		public float round2digits(float f) {
		    return	 Math.round(f * 100f) / 100f;
		}
	}

	public static void main(String[] args){
		if (args.length < 1) {
			System.out.println("USAGE : java -jar file.ecl");
		} else {
			String file = args[0];
			Benchmark ben = new Benchmark(file);
			ben.solveAll();
		}
	}
}
